/////////////////////////////////////////////////////////////////////
//
// winmain.cpp : entry-point and wndproc
//
/////////////////////////////////////////////////////////////////////

#include	"stdafx.h"
#include	"resource.h"



//	DEFINES		/////////////////////////////////////////////////////

#define			TRAYICON_ID			1
#define			WM_USER_TRAY		(WM_USER + 0x0100)

#define			TIMER_ID			1
#define			TIMER_SPAN			50				// ms

#define			IDM_UNUSED			8000			// must be unique
#define			IDM_TRACKBASE		8010			// must be unique

#define			MCIFAILED(ret)		(ret)

#define			HANDLE_DLGMSG(message, procedure) \
					case (message): (procedure(hDlg)); return TRUE
#define			HANDLE_DLGCMD(message, procedure) \
					case (message): (procedure(hDlg)); break



//	GLOBALS		/////////////////////////////////////////////////////

HINSTANCE		hInst;
HMENU			hMenu = NULL;

RECT			rtDlgSize;

MCIDEVICEID		wDeviceID;
int				iTracks = 0;
bool			fDevOpened = FALSE;



//	FUNCTIONS	/////////////////////////////////////////////////////

// tray icon manipulation : add
bool TrayIcon_Add( HWND hDlg, UINT uIconID, char* sText )
{
	HICON hIcon;
	NOTIFYICONDATA tnid;

	hIcon = LoadIcon( hInst, MAKEINTRESOURCE(uIconID) );

	ZeroMemory( &tnid, sizeof(NOTIFYICONDATA) );
	tnid.cbSize				= sizeof(NOTIFYICONDATA);
	tnid.uFlags				= NIF_ICON | NIF_MESSAGE | NIF_TIP;
	tnid.hWnd				= hDlg;
	tnid.hIcon				= hIcon;
	tnid.uID				= TRAYICON_ID;
	tnid.uCallbackMessage	= WM_USER_TRAY;

	strcpy( tnid.szTip, sText );

	// notify
	if( !Shell_NotifyIcon(NIM_ADD, &tnid) )
		return FALSE;

	return TRUE;
}


// tray icon manipulation : modify
bool TrayIcon_Modify( HWND hDlg, UINT uIconID = NULL, char* sText = NULL )
{
	HICON hIcon;
	NOTIFYICONDATA tnid;

	ZeroMemory( &tnid, sizeof(NOTIFYICONDATA) );
	tnid.cbSize		= sizeof(NOTIFYICONDATA);
	tnid.uFlags		= 0;
	tnid.hWnd		= hDlg;
	tnid.uID		= TRAYICON_ID;

	if( uIconID )
	{
		hIcon = LoadIcon( hInst, MAKEINTRESOURCE(uIconID) );
		tnid.hIcon	= hIcon;
		tnid.uFlags	|= NIF_ICON;
	}

	if( sText )
	{
		strcpy( tnid.szTip, sText );
		tnid.uFlags	|= NIF_TIP;
	}

	// notify
	if( !Shell_NotifyIcon(NIM_MODIFY, &tnid) )
		return FALSE;

	return TRUE;
}


// tray icon manipulation : delete
bool TrayIcon_Delete( HWND hDlg )
{
	NOTIFYICONDATA tnid;

	ZeroMemory( &tnid, sizeof(NOTIFYICONDATA) );
	tnid.cbSize				= sizeof(NOTIFYICONDATA);
	tnid.uFlags				= 0;
	tnid.hWnd				= hDlg;
	tnid.uID				= TRAYICON_ID;

	// notify
	if( !Shell_NotifyIcon(NIM_DELETE, &tnid) )
		return FALSE;

	return TRUE;
}


// is device that state; for instance, MCI_MODE_PLAY, MCI_MODE_PAUSE
bool DeviceIsState( DWORD dwMode )
{
	// check the state
	MCI_STATUS_PARMS mciStatus;
	mciStatus.dwItem = MCI_STATUS_MODE;

	mciSendCommand( wDeviceID, MCI_STATUS,
		MCI_STATUS_ITEM, (DWORD)&mciStatus );

	return mciStatus.dwReturn == dwMode;
}


// update tray-icon
void UpdateTrayIcon( HWND hDlg )
{
	char sTitle[128], sState[32];
	GetWindowText( hDlg, sTitle, sizeof(sTitle) );

	if( !fDevOpened )
		TrayIcon_Modify( hDlg, IDI_TRAY, NULL );

	if( DeviceIsState(MCI_MODE_PLAY) )
	{
		LoadString( hInst, IDS_TIP_PLAY, sState, sizeof(sState) );
		strcat( sTitle, sState );
		TrayIcon_Modify( hDlg, IDI_TRAY_PLAY, sTitle );
	}
	else
	{
		LoadString( hInst, IDS_TIP_STOP, sState, sizeof(sState) );
		strcat( sTitle, sState );
		TrayIcon_Modify( hDlg, IDI_TRAY_STOP, sTitle );
	}
}


// clean up cd-drive
void DeviceCleanup( HWND hDlg )
{
	if( fDevOpened )
		mciSendCommand( wDeviceID, MCI_CLOSE, 0, NULL );

	fDevOpened = FALSE;
	TrayIcon_Modify( hDlg, IDI_TRAY, NULL );
}


// enumerate tracks
bool MenuEnumTracks( HWND hDlg )
{
	DWORD dwRet;

	// menu (re)structuring
	char sText[128];
	MENUITEMINFO mti;

	ZeroMemory( &mti, sizeof(MENUITEMINFO) );
	mti.cbSize		= sizeof(MENUITEMINFO);
	mti.fMask		= MIIM_TYPE | MIIM_ID | MIIM_STATE;
	mti.fType		= MFT_STRING;
	mti.dwTypeData	= sText;
	mti.cch			= sizeof(sText);

	// if failed
	if( !fDevOpened )
	{
		mti.fState	= MFS_GRAYED;
		mti.wID		= IDM_UNUSED;

		LoadString( hInst, IDS_NODEV, sText, sizeof(sText) );
		InsertMenuItem( hMenu, 0, TRUE, &mti );
		TrayIcon_Modify( hDlg, IDI_TRAY, sText );

		return FALSE;
	}

	// set the time format to MSF
	MCI_SET_PARMS mciSet;
	mciSet.dwTimeFormat = MCI_FORMAT_MSF;

	dwRet = mciSendCommand( wDeviceID, MCI_SET, MCI_SET_TIME_FORMAT,
		(DWORD)&mciSet );

	if( MCIFAILED(dwRet) )
		return ( DeviceCleanup(hDlg), FALSE );

	// get number of tracks
	MCI_STATUS_PARMS mciStatus;
    mciStatus.dwItem = MCI_STATUS_NUMBER_OF_TRACKS;

	dwRet = mciSendCommand( wDeviceID, MCI_STATUS, MCI_STATUS_ITEM,
		(DWORD)&mciStatus );

	if( MCIFAILED(dwRet) )
	{
		// failed
		mti.fState	= MFS_GRAYED;
		mti.wID		= IDM_UNUSED;

		LoadString( hInst, IDS_NOTRACK, sText, sizeof(sText) );
		InsertMenuItem( hMenu, 0, TRUE, &mti );
		TrayIcon_Modify( hDlg, IDI_TRAY, sText );

		DeviceCleanup(hDlg);

		return FALSE;
	}

	iTracks = mciStatus.dwReturn;

	// build menu
	char sForm[128], sFormErr[128];
	int i;
	bool fAudio;

	LoadString( hInst, IDS_FORMAT, sForm, sizeof(sForm) );
	LoadString( hInst, IDS_FORMAT_NOTAUDIO, sFormErr, sizeof(sFormErr) );

	mti.fType = MFT_STRING | MFT_RADIOCHECK;

	for( i = 0; i < iTracks; i++ )
	{
		// is audio?
		mciStatus.dwTrack = i + 1;
		mciStatus.dwItem = MCI_CDA_STATUS_TYPE_TRACK;
		dwRet = mciSendCommand( wDeviceID, MCI_STATUS,
			MCI_STATUS_ITEM | MCI_TRACK, (DWORD)&mciStatus );

		fAudio = !MCIFAILED( dwRet ) &&
			( mciStatus.dwReturn == MCI_CDA_TRACK_AUDIO );

		// get track info
		mciStatus.dwItem = MCI_STATUS_LENGTH;
		dwRet = mciSendCommand( wDeviceID, MCI_STATUS,
			MCI_STATUS_ITEM | MCI_TRACK, (DWORD)&mciStatus );

		if( MCIFAILED(dwRet) )
		{
			// failed
			mti.fState	= MFS_GRAYED;
			mti.wID		= IDM_UNUSED;

			LoadString( hInst, IDS_ERROR, sText, sizeof(sText) );
			InsertMenuItem( hMenu, 0, TRUE, &mti );
			TrayIcon_Modify( hDlg, IDI_TRAY, sText );

			DeviceCleanup( hDlg );

			return FALSE;
		}

		// format menu-item
		if( fAudio )
			sprintf( sText, sForm,
				i + 1,
				MCI_MSF_MINUTE(mciStatus.dwReturn),
				MCI_MSF_SECOND(mciStatus.dwReturn) );
		else
			sprintf( sText, sFormErr, i + 1 );

		mti.fState = fAudio ? 0 : MFS_GRAYED;

		// insert
		mti.wID	= IDM_TRACKBASE + i;

		InsertMenuItem( hMenu, 0, TRUE, &mti );
	}

	return TRUE;
}


// init menu
bool MenuInitialize( HWND hDlg )
{
	// menu initialize
	if( hMenu )
		DestroyMenu( hMenu );
	hMenu = GetSubMenu( LoadMenu(hInst, MAKEINTRESOURCE(IDRM_POPUP)), 0 );

	// set default menu-item
	MENUITEMINFO mti;
	ZeroMemory( &mti, sizeof(MENUITEMINFO) );
	mti.cbSize	= sizeof(MENUITEMINFO);
	mti.fMask	= MIIM_STATE;
	mti.fState	= MFS_DEFAULT;

	SetMenuItemInfo( hMenu, IDM_PLAYPAUSE, FALSE, &mti );

	// enum tracks
	if( !MenuEnumTracks(hDlg) )
		return FALSE;

	// update icon
	UpdateTrayIcon( hDlg );

	return TRUE;
}


// init cd-drive
bool DeviceInitialize( HWND hDlg )
{
	DWORD dwRet;

	// close
	DeviceCleanup( hDlg );

	// menu (re)structuring
	MENUITEMINFO mti;
	char sText[128];

	ZeroMemory( &mti, sizeof(MENUITEMINFO) );
	mti.cbSize		= sizeof(MENUITEMINFO);
	mti.fMask		= MIIM_TYPE | MIIM_ID | MIIM_STATE;
	mti.fType		= MFT_STRING;
	mti.dwTypeData	= sText;
	mti.cch			= sizeof(sText);
	mti.fState		= 0;

	// open device
	MCI_OPEN_PARMS mciOpen;
	mciOpen.lpstrDeviceType = (LPSTR)MCI_DEVTYPE_CD_AUDIO;

	dwRet = mciSendCommand( NULL, MCI_OPEN,
		MCI_OPEN_TYPE | MCI_OPEN_TYPE_ID, (DWORD)&mciOpen );

	if( MCIFAILED(dwRet) )
		fDevOpened = FALSE;
	else
	{
		fDevOpened = TRUE;
		wDeviceID = mciOpen.wDeviceID;
	}

	// build menu
	return MenuInitialize( hDlg );
}


// play a track
bool DevicePlay( HWND hDlg, int iTrack )
{
	DWORD dwRet;

	// check init
	if( !fDevOpened )
		return ( DeviceInitialize(hDlg), FALSE );

	// set the time format to TMSF
	MCI_SET_PARMS mciSet;
	mciSet.dwTimeFormat = MCI_FORMAT_TMSF;

	dwRet = mciSendCommand( wDeviceID, MCI_SET, MCI_SET_TIME_FORMAT,
		(DWORD)&mciSet );

	if( MCIFAILED(dwRet) )
		return ( DeviceInitialize(hDlg), FALSE );

	// play a track
	MCI_PLAY_PARMS mciPlay;
	mciPlay.dwFrom = MCI_MAKE_TMSF( (BYTE)iTrack, 0, 0, 0 );

	dwRet = mciSendCommand( wDeviceID, MCI_PLAY, MCI_FROM,
		(DWORD)&mciPlay );

	if( MCIFAILED(dwRet) )
		return ( DeviceInitialize(hDlg), FALSE );

	// update icon
	UpdateTrayIcon( hDlg );

	return TRUE;
}


// stop
bool DeviceStop( HWND hDlg )
{
	// check init
	if( !fDevOpened )
		return ( DeviceInitialize(hDlg), FALSE );

	DWORD dwRet;
	dwRet = mciSendCommand( wDeviceID, MCI_STOP, 0, NULL );

	if( MCIFAILED(dwRet) )
		return ( DeviceInitialize(hDlg), FALSE );

	// update icon
	UpdateTrayIcon( hDlg );

	return TRUE;
}


// play/pause
bool DevicePlayPause( HWND hDlg )
{
	// check init
	if( !fDevOpened )
		return ( DeviceInitialize(hDlg), FALSE );

	DWORD dwRet;
	dwRet = mciSendCommand( wDeviceID,
		DeviceIsState(MCI_MODE_PLAY) ? MCI_PAUSE : MCI_PLAY, 0, NULL );

	if( MCIFAILED(dwRet) )
		return ( DeviceInitialize(hDlg), FALSE );

	// update icon
	UpdateTrayIcon( hDlg );

	return TRUE;
}


// tray close
bool DeviceTrayClose( HWND hDlg )
{
	// check init
	if( !fDevOpened )
		return ( DeviceInitialize(hDlg), FALSE );

	DWORD dwRet;
	dwRet = mciSendCommand( wDeviceID, MCI_SET, MCI_SET_DOOR_CLOSED, NULL );

	if( MCIFAILED(dwRet) )
		return ( DeviceInitialize(hDlg), FALSE );

	// update
	UpdateTrayIcon( hDlg );
	MenuInitialize( hDlg );

	return TRUE;
}


// tray open
bool DeviceTrayOpen( HWND hDlg )
{
	// check init
	if( !fDevOpened )
		return ( DeviceInitialize(hDlg), FALSE );

	DWORD dwRet;
	dwRet = mciSendCommand( wDeviceID, MCI_SET, MCI_SET_DOOR_OPEN, NULL );

	if( MCIFAILED(dwRet) )
		return ( DeviceInitialize(hDlg), FALSE );

	// update icon
	UpdateTrayIcon( hDlg );

	return TRUE;
}


// menu popup
void MenuPopup( HWND hDlg )
{
	// is opened?
	if( !fDevOpened )
		DeviceInitialize( hDlg );
	else
	{
		// get current track
		MCI_STATUS_PARMS mciStatus;
		mciStatus.dwItem = MCI_STATUS_CURRENT_TRACK;

		mciSendCommand( wDeviceID, MCI_STATUS,
			MCI_STATUS_ITEM, (DWORD)&mciStatus );

		// menu
		MENUITEMINFO mti;
		char sText[128];
		int i;

		ZeroMemory( &mti, sizeof(MENUITEMINFO) );
		mti.cbSize	= sizeof(MENUITEMINFO);
		mti.fMask	= MIIM_STATE;

		// uncheck all
		for( i = 0; i < iTracks; i++ )
		{
			GetMenuItemInfo( hMenu, IDM_TRACKBASE + i, FALSE, &mti );
			mti.fState &= ~MFS_CHECKED;

			SetMenuItemInfo( hMenu, IDM_TRACKBASE + i, FALSE, &mti );
		}

		// check the current track if playing
		if( !DeviceIsState(MCI_MODE_STOP) )
		{
			GetMenuItemInfo( hMenu, IDM_TRACKBASE + i, FALSE, &mti );
			mti.fState |= MFS_CHECKED;

			SetMenuItemInfo( hMenu, IDM_TRACKBASE + mciStatus.dwReturn - 1,
				FALSE, &mti );
		}

		// set menu-item text to "play" or "pause"
		mti.fMask		= MIIM_TYPE;
		mti.fType		= MFT_STRING;
		mti.dwTypeData	= sText;
		mti.cch			= sizeof(sText);

		LoadString( hInst,
			DeviceIsState(MCI_MODE_PLAY) ? IDS_MENU_PAUSE : IDS_MENU_PLAY,
			sText, sizeof(sText) );

		SetMenuItemInfo( hMenu, IDM_PLAYPAUSE, FALSE, &mti );
	}

	// popup
	POINT pt;
	GetCursorPos( &pt );

	SetForegroundWindow( hDlg );
	TrackPopupMenu( hMenu,
		TPM_RIGHTALIGN | TPM_BOTTOMALIGN | TPM_RIGHTBUTTON,
		pt.x, pt.y, 0,
		hDlg, NULL );
}


// dialog init
void DlgInitialize( HWND hDlg )
{
	// hide window; zero-size
	GetWindowRect( hDlg, &rtDlgSize );
	SetWindowPos( hDlg, NULL, 0, 0, 0, 0,
		SWP_NOZORDER | SWP_HIDEWINDOW | SWP_NOACTIVATE );

	// tray icon
	char sTitle[128];
	GetWindowText( hDlg, sTitle, sizeof(sTitle) );
	TrayIcon_Add( hDlg, IDI_TRAY, sTitle );

	// init dev
	DeviceInitialize( hDlg );

	// init timer
	SetTimer( hDlg, TIMER_ID, TIMER_SPAN, NULL );
}


// dialog clean up
void DlgCleanup( HWND hDlg )
{
	DeviceCleanup( hDlg );
	TrayIcon_Delete( hDlg );

	EndDialog( hDlg, 0 );
}


// dialog loop
BOOL WINAPI DlgProc( HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam )
{
	UINT id;

	switch( msg )
	{
		HANDLE_DLGMSG( WM_INITDIALOG, DlgInitialize );

		case WM_TIMER:
			// timer runs once
			KillTimer( hDlg, TIMER_ID );

			// restore window size
			SetWindowPos( hDlg, NULL,
				rtDlgSize.left, rtDlgSize.top,
				rtDlgSize.right - rtDlgSize.left, rtDlgSize.bottom - rtDlgSize.top,
				SWP_NOZORDER | SWP_HIDEWINDOW | SWP_NOACTIVATE );
			return TRUE;

		case WM_COMMAND:
			// command id
			id = LOWORD(wParam);

			// track play command
			if( id - IDM_TRACKBASE < 100 )
				DevicePlay( hDlg, id - IDM_TRACKBASE + 1 );

			// other commands
			else switch( id )
			{
				// show/hide
				case IDOK:
					ShowWindow( hDlg, SW_HIDE );
					break;

				case IDM_SETTING:
					ShowWindow( hDlg, SW_SHOW );
					break;

				// exit
				case IDM_EXITSTOP:
					DeviceStop( hDlg );
				case IDM_EXIT:
				HANDLE_DLGCMD( IDCANCEL, DlgCleanup );

				// general commands
				HANDLE_DLGCMD( IDM_REFRESH, DeviceInitialize );
				HANDLE_DLGCMD( IDM_STOP, DeviceStop );
				HANDLE_DLGCMD( IDM_PLAYPAUSE, DevicePlayPause );
				HANDLE_DLGCMD( IDM_DOORCLOSE, DeviceTrayClose );
				HANDLE_DLGCMD( IDM_DOOROPEN, DeviceTrayOpen );

				default:
					return FALSE;
			}
			return TRUE;

		case WM_USER_TRAY:
			switch( lParam )
			{
				HANDLE_DLGMSG( WM_LBUTTONDOWN, DevicePlayPause );
				HANDLE_DLGMSG( WM_RBUTTONDOWN, MenuPopup );
			}
			break;
	}
	return FALSE;
}


// entry point
int APIENTRY WinMain( HINSTANCE hInstance, HINSTANCE, LPSTR, int )
{
	// globals
	hInst = hInstance;

	// prevent

	// dialog
	return DialogBoxParam(
		hInst, MAKEINTRESOURCE(IDD_MAIN), NULL, (DLGPROC)DlgProc, 0 );
}







